@@ -3,6 +3,7 @@
 
 import json
 
+from collections import OrderedDict
 from copy import deepcopy
 
 from formiodata.builder import Builder
@@ -42,18 +43,22 @@ class Form:
         self.components = {}
         self.load_components()
         self.data = FormData(self)
-        self.renderer = FormRenderer(self)
+
+    def render(self):
+        return FormRenderer(self)
 
     def set_builder_by_builder_schema_json(self):
         self.builder = Builder(self.builder_schema_json, self.lang)
 
     def load_components(self):
+        # TODO: Make recursive(?)
         for key, component in self.builder.form_components.items():
             # Rather lazy check, but sane.
             # if not self.form.get(key):
             #     continue
             # replaced with default value and fast
             raw_value = self.form.get(key, component.defaultValue)
+            component.component_owner = self.builder
             component.value = raw_value
             component.raw_value = raw_value
             self.components[key] = component
@@ -69,23 +74,21 @@ class FormRenderer:
     def __init__(self, form):
         self.form = form
         self.builder = form.builder
-        self.components = []
-        self.component_ids = {}
         self.load_components()
 
     def load_components(self):
         """ Loads the components (tree) to render, with values
         (data) and obtaining the tree by creating all sub-components
         e.g. in layout and datagrid. """
-        for component in self.builder.raw_components:
-            # Only determine and load class if component type.
-            if 'type' in component:
-                builder_component_obj = self.builder.component_ids[component['id']]
-                # New object, don't affect the Builder component
-                component_obj = self.builder.get_component_object(builder_component_obj.raw)
-                component_obj.load(None, self.form.form, renderer=self)
-                self.components.append(component_obj)
+        self.builder.load_components()
 
+    @property
+    def components(self):
+        return self.builder.components
+
+    @property
+    def component_ids(self):
+        return self.builder.component_ids
 
 class FormData:
 
